@@ -1,43 +1,23 @@
 #!/usr/bin/python3
 
-import apt
 import codecs
 import fnmatch
 import gettext
 import os
-import platform
 import re
 import sys
-from html.parser import HTMLParser
 import traceback
+from html.parser import HTMLParser
 
+import apt
 from gi.repository import Gio
 
-from Classes import Update, Alias, Rule, KERNEL_PKG_NAMES
+from Classes import (CONFIGURED_KERNEL_TYPE, KERNEL_PKG_NAMES,
+                     PRIORITY_UPDATES, SUPPORTED_KERNEL_TYPES, Alias, KernelVersion, Update)
 
 gettext.install("mintupdate", "/usr/share/locale")
 
-class KernelVersion():
-
-    def __init__(self, version):
-        self.version = version
-        version_array = self.version.replace("-", ".").split(".")
-        self.numeric_versions = []
-        for i in range(4):
-            element = version_array[i]
-            if len(element) == 1:
-                element = "00%s" % element
-            elif len(element) == 2:
-                element = "0%s" % element
-            self.numeric_versions.append(element)
-        self.numeric_representation = ".".join(self.numeric_versions)
-        self.std_version = "%s.%s.%s-%s" % (version_array[0], version_array[1], version_array[2], version_array[3])
-        self.series = "%s.%s.%s" % (version_array[0], version_array[1], version_array[2])
-
-# These updates take priority over other updates.
-# If a new version of these packages is available,
-# nothing else is listed.
-PRIORITY_UPDATES = ['mintupdate', 'mint-upgrade-info']
+meta_names = []
 
 class APTCheck():
 
@@ -45,7 +25,6 @@ class APTCheck():
         self.settings = Gio.Settings("com.linuxmint.updates")
         self.cache = apt.Cache()
         self.priority_updates_available = False
-        self.load_rules()
 
     def load_aliases(self):
         self.aliases = {}
@@ -60,40 +39,8 @@ class APTCheck():
                             alias_package = alias_package.strip()
                             self.aliases[alias_package] = alias_object
 
-    def load_rules(self):
-        self.named_rules = {}
-        self.wildcard_rules = {}
-        with open("/usr/lib/linuxmint/mintUpdate/rules","r") as rulesFile:
-            rules = rulesFile.readlines()
-            for rule in rules:
-                if "|" not in rule:
-                    continue
-                rule_fields = rule.strip().split("|")
-                if (len(rule_fields) == 3):
-                    rule_level = int(rule_fields[0])
-                    rule_name = rule_fields[1]
-                    rule_version = rule_fields[2]
-                    rule = Rule(rule_name, rule_version, rule_level)
-                    if rule.is_wildcard:
-                        self.wildcard_rules[rule.name] = rule
-                    else:
-                        self.named_rules[rule.name] = rule
-
-    def refresh_cache(self):
-        if os.getuid() == 0 :
-            if "--use-synaptic" in sys.argv:
-                window_id = int(sys.argv[2])
-                from subprocess import Popen
-                cmd = ["sudo", "/usr/sbin/synaptic", "--hide-main-window", "--update-at-startup", "--non-interactive", "--parent-window-id", "%d" % window_id]
-                comnd = Popen(' '.join(cmd), shell=True)
-                comnd.wait()
-            else:
-                self.cache.update()
-
     def find_changes(self):
-        # Reopen the cache to reflect any updates
-        self.cache.open(None)
-        self.cache.upgrade(self.settings.get_boolean("dist-upgrade"))
+        self.cache.upgrade(True) # dist-upgrade
         changes = self.cache.get_changes()
 
         self.updates = {}
@@ -104,131 +51,189 @@ class APTCheck():
                 self.add_update(pkg)
 
         # Kernel updates
-        if self.settings.get_boolean("use-lowlatency-kernels"):
-            kernel_type = "-lowlatency"
-        else:
-            kernel_type = "-generic"
-
-        meta_names = []
-        _metas = [s for s in self.cache.keys() if s.startswith("linux" + kernel_type)]
+        lts_meta_name = "linux" + CONFIGURED_KERNEL_TYPE
+        _metas = [s for s in self.cache.keys() if s.startswith(lts_meta_name)]
+        if CONFIGURED_KERNEL_TYPE == "-generic":
+            _metas.append("linux-virtual")
+        global meta_names
         for meta in _metas:
             shortname = meta.split(":")[0]
             if shortname not in meta_names:
                 meta_names.append(shortname)
-
         try:
-            if self.settings.get_boolean("kernel-updates-are-visible"):
+            # Get the uname version
+            active_kernel = KernelVersion(os.uname().release)
 
-                # Get the uname version
-                uname_kernel = KernelVersion(platform.release())
+            # Override installed kernel if not of the configured type
+            try:
+                active_kernel_type = "-" + active_kernel.version.split("-")[-1]
+            except:
+                active_kernel_type = CONFIGURED_KERNEL_TYPE
+            if  active_kernel_type != CONFIGURED_KERNEL_TYPE:
+                active_kernel.series = ("0","0","0")
 
-                # Check if any meta is installed..
-                meta_installed = False
-                for meta_name in meta_names:
-                    if meta_name in self.cache:
-                        meta = self.cache[meta_name]
+            # Uncomment for testing:
+            # active_kernel = KernelVersion("4.18.0-0-generic")
+
+            # Check if any meta is installed..
+            meta_candidate_same_series = None
+            meta_candidate_higher_series = None
+            for meta_name in meta_names:
+                if meta_name in self.cache:
+                    meta = self.cache[meta_name]
+                    meta_kernel = KernelVersion(meta.candidate.version)
+                    if (active_kernel.series > meta_kernel.series):
+                        # Meta is lower than the installed kernel series, ignore
+                        continue
+                    else:
                         if meta.is_installed:
-                            meta_installed = True
+                            # Meta is already installed, return
+                            return
+                        # never install linux-virtual, we only support it being
+                        # installed already
+                        if meta_name == "linux-virtual":
+                            continue
+                        # Meta is not installed, make it a candidate if higher
+                        # than any current candidate
+                        if active_kernel.series == meta_kernel.series:
+                            # same series
+                            if (not meta_candidate_same_series or meta_kernel.version_id >
+                                KernelVersion(meta_candidate_same_series.candidate.version).version_id
+                                ):
+                                meta_candidate_same_series = meta
+                        else:
+                            # higher series
+                            if (not meta_candidate_higher_series or meta_kernel.version_id >
+                                KernelVersion(meta_candidate_higher_series.candidate.version).version_id
+                                ):
+                                meta_candidate_higher_series = meta
+
+            # If we're here, no meta was installed
+            if meta_candidate_same_series:
+                # but a candidate of the same series was found, add to updates and return
+                self.add_update(meta_candidate_same_series, kernel_update=True)
+                return
+
+            # If we're here, no matching meta was found
+            if meta_candidate_higher_series:
+                # but we found a higher meta candidate, add it to the list of updates
+                # unless the installed kernel series is lower than the LTS series
+                # for some reason, in the latter case force the LTS meta
+                if meta_candidate_higher_series.name != lts_meta_name:
+                    if lts_meta_name in self.cache:
+                        lts_meta = self.cache[lts_meta_name]
+                        lts_meta_kernel = KernelVersion(lts_meta.candidate.version)
+                        if active_kernel.series < lts_meta_kernel.series:
+                            meta_candidate_higher_series = lts_meta
+                self.add_update(meta_candidate_higher_series, kernel_update=True)
+                return
+
+            # We've gone past all the metas, so we should recommend the latest
+            # kernel on the series we're in
+            max_kernel = active_kernel
+            for pkgname in self.cache.keys():
+                match = re.match(r'^(?:linux-image-)(\d.+?)%s$' % active_kernel_type, pkgname)
+                if match:
+                    kernel = KernelVersion(match.group(1))
+                    if kernel.series == max_kernel.series and kernel.version_id > max_kernel.version_id:
+                        max_kernel = kernel
+            if max_kernel.version_id != active_kernel.version_id:
+                for pkgname in KERNEL_PKG_NAMES:
+                    pkgname = pkgname.replace('VERSION', max_kernel.version).replace("-KERNELTYPE", active_kernel_type)
+                    if pkgname in self.cache:
+                        pkg = self.cache[pkgname]
+                        if not pkg.is_installed:
+                            self.add_update(pkg, kernel_update=True)
                             return
 
-                # If no meta is installed, try to recommend one
-                if not meta_installed:
-                    for meta_name in meta_names:
-                        if meta_name in self.cache:
-                            meta = self.cache[meta_name]
-                            recommended_kernel = KernelVersion(meta.candidate.version)
-                            if (uname_kernel.numeric_representation <= recommended_kernel.numeric_representation):
-                                # This meta version is >= to the uname version, add it as an update
-                                self.add_update(meta, kernel_update=True)
-                                # Return because we only want to add one meta, and we don't want to recommend latest kernels in the series
-                                return
+        except:
+            traceback.print_exc()
 
-                # We've gone past all the metas, so we should recommend the latest kernel on the series we're in
-                max_kernel = uname_kernel
-                kernel_type = "-generic"
-                if self.settings.get_boolean("use-lowlatency-kernels"):
-                    kernel_type = "-lowlatency"
-                for pkgname in self.cache.keys():
-                    match = re.match(r'^(?:linux-image-)(?:unsigned-)?(\d.+?)' + kernel_type + '$', pkgname)
-                    if match:
-                        kernel = KernelVersion(match.group(1))
-                        if kernel.numeric_representation > max_kernel.numeric_representation and kernel.series == max_kernel.series:
-                            max_kernel = kernel
-                if max_kernel.numeric_representation != uname_kernel.numeric_representation:
-                    for pkgname in KERNEL_PKG_NAMES:
-                        pkgname = pkgname.replace('VERSION', max_kernel.std_version)
-                        if pkgname in self.cache:
-                            pkg = self.cache[pkgname]
-                            if not pkg.is_installed:
-                                self.add_update(pkg, kernel_update=True)
+    def is_blacklisted(self, source_name, version):
+        for blacklist in self.settings.get_strv("blacklisted-packages"):
+            if "=" in blacklist:
+                (bl_pkg, bl_ver) = blacklist.split("=", 1)
+            else:
+                bl_pkg = blacklist
+                bl_ver = None
+            if fnmatch.fnmatch(source_name, bl_pkg) and (not bl_ver or bl_ver == version):
+                return True
+        return False
 
-        except Exception:
-            print(sys.exc_info()[0])
+    def get_kernel_version_from_meta_package(self, pkg):
+        for dependency in pkg.dependencies:
+            if not dependency.target_versions or not dependency.rawtype == "Depends":
+                return None
+            deppkg = dependency.target_versions[0]
+            if deppkg.source_name in ("linux", "linux-signed"):
+                return deppkg.source_version
+            if deppkg.source_name.startswith("linux-meta"):
+                return self.get_kernel_version_from_meta_package(deppkg)
+        return None
 
     def add_update(self, package, kernel_update=False):
+        source_version = package.candidate.version
+        # Change version of kernel meta packages to that of the actual kernel
+        # for grouping with related updates
+        if package.candidate.source_name.startswith("linux-meta"):
+            _source_version = self.get_kernel_version_from_meta_package(package.candidate)
+            if _source_version:
+                source_version = _source_version
 
-        if package.name in ['linux-libc-dev', 'linux-kernel-generic']:
-            source_name = package.name
-        elif package.candidate.source_name in ['linux', 'linux-meta', 'linux-hwe', 'linux-hwe-edge']:
-            source_name = "linux-%s" % package.candidate.version
-        elif package.name.startswith("linux-image") or package.name.startswith("linux-headers") or package.name.startswith("linux-modules") or package.name.startswith("linux-tools"):
-            source_name = "linux-%s" % package.candidate.version
+        # Change source name of kernel packages for grouping with related updates
+        if (package.candidate.source_name == "linux" or
+            package.candidate.source_name.startswith("linux-hwe") or
+            package.candidate.source_name.startswith("linux-meta") or
+            package.candidate.source_name.startswith("linux-signed") or
+            [True for flavor in SUPPORTED_KERNEL_TYPES if package.candidate.source_name.startswith("linux%s" % flavor)]
+           ):
+            kernel_update = True
+            source_name = "linux-%s" % source_version
         else:
             source_name = package.candidate.source_name
 
         # ignore packages blacklisted by the user
-        for blacklist in self.settings.get_strv("blacklisted-packages"):
-            if fnmatch.fnmatch(source_name, blacklist):
-                return
+        if self.is_blacklisted(package.candidate.source_name, package.candidate.version):
+            return
 
         if source_name in PRIORITY_UPDATES:
-            if self.priority_updates_available == False and len(self.updates) > 0:
-                self.updates = {}
+            if not self.priority_updates_available and len(self.updates) > 0:
+                self.updates.clear()
             self.priority_updates_available = True
-        if (source_name in PRIORITY_UPDATES) or self.priority_updates_available == False:
+        if source_name in PRIORITY_UPDATES or not self.priority_updates_available:
             if source_name in self.updates:
                 update = self.updates[source_name]
                 update.add_package(package)
+                # Adjust update.old_version for kernel updates to try and
+                # match the kernel, not the meta
+                if kernel_update and package.is_installed and not \
+                   "-" in update.old_version and "-" in package.installed.version:
+                    update.old_version = package.installed.version
             else:
                 update = Update(package, source_name=source_name)
-
-                if source_name in self.named_rules.keys():
-                    rule = self.named_rules[source_name]
-                    if (rule.match(source_name, update.new_version)):
-                        update.level = rule.level
-                else:
-                    for rule_name in self.wildcard_rules.keys():
-                        rule = self.wildcard_rules[rule_name]
-                        if rule.match(source_name, update.new_version):
-                            update.level = rule.level
-                            break
-
-                if update.level < 5:
-                    self.updates[source_name] = update
+                self.updates[source_name] = update
             if kernel_update:
                 update.type = "kernel"
+            update.new_version = source_version
 
     def serialize_updates(self):
         # Print updates
-        for source_name in sorted(self.updates.keys()):
-            update = self.updates[source_name]
-            update.serialize()
-
-    def list_updates(self):
-        # Print updates
-        for source_name in sorted(self.updates.keys()):
+        for source_name in self.updates.keys():
             update = self.updates[source_name]
             update.serialize()
 
     def apply_aliases(self):
         for source_name in self.updates.keys():
             update = self.updates[source_name]
-            if source_name in self.aliases.keys():
+            if source_name in self.aliases:
                 alias = self.aliases[source_name]
                 update.display_name = alias.name
                 update.short_description = alias.short_description
                 update.description = alias.description
-            elif update.type == "kernel" and source_name not in ['linux-libc-dev', 'linux-kernel-generic']:
+            elif (update.type == "kernel" and
+                  source_name not in ['linux-libc-dev', 'linux-kernel-generic'] and
+                  (len(update.package_names) >= 3 or update.package_names[0] in meta_names)
+                 ):
                 update.display_name = _("Linux kernel %s") % update.new_version
                 update.short_description = _("The Linux kernel.")
                 update.description = _("The Linux Kernel is responsible for hardware and drivers support. Note that this update will not remove your existing kernel. You will still be able to boot with the current kernel by choosing the advanced options in your boot menu. Please be cautious though.. kernel regressions can affect your ability to connect to the Internet or to log in graphically. DKMS modules are compiled for the most recent kernels installed on your computer. If you are using proprietary drivers and you want to use an older kernel, you will need to remove the new one first.")
@@ -250,7 +255,7 @@ class APTCheck():
                     if line.startswith("Package: "):
                         try:
                             pkgname = line.replace("Package: ", "")
-                            if pkgname in self.updates.keys():
+                            if pkgname in self.updates:
                                 update = self.updates[pkgname]
                                 j = 2 # skip md5 line after package name line
                                 while True:
@@ -350,14 +355,13 @@ class APTCheck():
 if __name__ == "__main__":
     try:
         check = APTCheck()
-        check.refresh_cache()
         check.find_changes()
         check.apply_l10n_descriptions()
         check.load_aliases()
         check.apply_aliases()
         check.clean_descriptions()
         check.serialize_updates()
-        if os.path.exists("/usr/bin/mintinstall-update-pkgcache"):
+        if os.getuid() == 0 and os.path.exists("/usr/bin/mintinstall-update-pkgcache"):
             # Spawn the cache update asynchronously
             # We're using os.system with & here to make sure it's async and detached
             # from the caller (which will die before the child process is finished)
